/**
 * Copyright (C) 2012-2018 52°North Initiative for Geospatial Open Source
 * Software GmbH
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 as published
 * by the Free Software Foundation.
 *
 * If the program is linked with libraries which are licensed under one of
 * the following licenses, the combination of the program with the linked
 * library is not considered a "derivative work" of the program:
 *
 *     - Apache License, version 2.0
 *     - Apache Software License, version 1.0
 *     - GNU Lesser General Public License, version 3
 *     - Mozilla Public License, versions 1.0, 1.1 and 2.0
 *     - Common Development and Distribution License (CDDL), version 1.0
 *
 * Therefore the distribution of the program linked with libraries licensed
 * under the aforementioned licenses, is permitted by the copyright holders
 * if the distribution is compliant with both the GNU General Public
 * License version 2 and the aforementioned licenses.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 */
package org.n52.sos.encode;

import java.util.Collections;
import java.util.EnumMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import org.apache.xmlbeans.XmlCursor;
import org.apache.xmlbeans.XmlException;
import org.apache.xmlbeans.XmlObject;
import org.isotc211.x2005.gmd.EXExtentType;
import org.n52.sos.exception.ows.NoApplicableCodeException;
import org.n52.sos.exception.ows.concrete.DateTimeFormatException;
import org.n52.sos.exception.ows.concrete.UnsupportedEncoderInputException;
import org.n52.sos.ogc.OGCConstants;
import org.n52.sos.ogc.gml.AbstractCRS;
import org.n52.sos.ogc.gml.AbstractCoordinateSystem;
import org.n52.sos.ogc.gml.AbstractDatum;
import org.n52.sos.ogc.gml.AbstractFeature;
import org.n52.sos.ogc.gml.AbstractGeometry;
import org.n52.sos.ogc.gml.CodeWithAuthority;
import org.n52.sos.ogc.gml.CoordinateSystemAxis;
import org.n52.sos.ogc.gml.DefaultEncoding;
import org.n52.sos.ogc.gml.Definition;
import org.n52.sos.ogc.gml.DomainOfValidity;
import org.n52.sos.ogc.gml.GenericMetaData;
import org.n52.sos.ogc.gml.GmlConstants;
import org.n52.sos.ogc.gml.VerticalCRS;
import org.n52.sos.ogc.gml.VerticalCS;
import org.n52.sos.ogc.gml.VerticalDatum;
import org.n52.sos.ogc.gml.time.Time;
import org.n52.sos.ogc.gml.time.Time.TimeIndeterminateValue;
import org.n52.sos.ogc.gml.time.TimeInstant;
import org.n52.sos.ogc.gml.time.TimePeriod;
import org.n52.sos.ogc.gml.time.TimePosition;
import org.n52.sos.ogc.om.features.FeatureCollection;
import org.n52.sos.ogc.om.features.samplingFeatures.AbstractSamplingFeature;
import org.n52.sos.ogc.om.values.CategoryValue;
import org.n52.sos.ogc.om.values.QuantityValue;
import org.n52.sos.ogc.ows.OwsExceptionReport;
import org.n52.sos.ogc.sos.SosConstants.HelperValues;
import org.n52.sos.ogc.sos.SosEnvelope;
import org.n52.sos.service.ServiceConfiguration;
import org.n52.sos.util.CodingHelper;
import org.n52.sos.util.DateTimeHelper;
import org.n52.sos.util.JTSHelper;
import org.n52.sos.util.JavaHelper;
import org.n52.sos.util.MinMax;
import org.n52.sos.util.OMHelper;
import org.n52.sos.util.SosHelper;
import org.n52.sos.util.XmlHelper;
import org.n52.sos.util.XmlOptionsHelper;
import org.n52.sos.w3c.Nillable;
import org.n52.sos.w3c.SchemaLocation;
import org.n52.sos.w3c.xlink.Reference;
import org.n52.sos.w3c.xlink.Referenceable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3.x1999.xlink.ActuateType;
import org.w3.x1999.xlink.ShowType;
import org.w3.x1999.xlink.TypeType;

import com.google.common.base.Joiner;
import com.google.common.base.Strings;
import com.google.common.collect.Sets;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.geom.util.PolygonExtracter;

import net.opengis.gml.x32.AbstractCRSType;
import net.opengis.gml.x32.AbstractCoordinateSystemType;
import net.opengis.gml.x32.AbstractDatumType;
import net.opengis.gml.x32.AbstractGeometryType;
import net.opengis.gml.x32.AbstractRingPropertyType;
import net.opengis.gml.x32.AbstractRingType;
import net.opengis.gml.x32.AggregationType;
import net.opengis.gml.x32.CodeType;
import net.opengis.gml.x32.CodeWithAuthorityType;
import net.opengis.gml.x32.CoordinateSystemAxisDocument;
import net.opengis.gml.x32.CoordinateSystemAxisPropertyType;
import net.opengis.gml.x32.CoordinateSystemAxisType;
import net.opengis.gml.x32.DefinitionType;
import net.opengis.gml.x32.DirectPositionListType;
import net.opengis.gml.x32.DirectPositionType;
import net.opengis.gml.x32.EnvelopeDocument;
import net.opengis.gml.x32.EnvelopeType;
import net.opengis.gml.x32.FeatureCollectionDocument;
import net.opengis.gml.x32.FeatureCollectionType;
import net.opengis.gml.x32.FeaturePropertyType;
import net.opengis.gml.x32.GenericMetaDataDocument;
import net.opengis.gml.x32.GenericMetaDataType;
import net.opengis.gml.x32.GeometryPropertyType;
import net.opengis.gml.x32.LineStringDocument;
import net.opengis.gml.x32.LineStringType;
import net.opengis.gml.x32.LinearRingType;
import net.opengis.gml.x32.MeasureType;
import net.opengis.gml.x32.MultiPointDocument;
import net.opengis.gml.x32.MultiPointType;
import net.opengis.gml.x32.PointDocument;
import net.opengis.gml.x32.PointType;
import net.opengis.gml.x32.PolygonDocument;
import net.opengis.gml.x32.PolygonType;
import net.opengis.gml.x32.ReferenceType;
import net.opengis.gml.x32.TimeIndeterminateValueType;
import net.opengis.gml.x32.TimeInstantDocument;
import net.opengis.gml.x32.TimeInstantPropertyType;
import net.opengis.gml.x32.TimeInstantType;
import net.opengis.gml.x32.TimePeriodDocument;
import net.opengis.gml.x32.TimePeriodPropertyType;
import net.opengis.gml.x32.TimePeriodType;
import net.opengis.gml.x32.TimePositionType;
import net.opengis.gml.x32.VerticalCRSPropertyType;
import net.opengis.gml.x32.VerticalCRSType;
import net.opengis.gml.x32.VerticalCSDocument;
import net.opengis.gml.x32.VerticalCSPropertyType;
import net.opengis.gml.x32.VerticalCSType;
import net.opengis.gml.x32.VerticalDatumDocument;
import net.opengis.gml.x32.VerticalDatumPropertyType;
import net.opengis.gml.x32.VerticalDatumType;

/**
 * @since 4.0.0
 * 
 */
public class GmlEncoderv321 extends AbstractGmlEncoderv321<Object> {

    private static final Logger LOGGER = LoggerFactory.getLogger(GmlEncoderv321.class);

    private static final Set<EncoderKey> ENCODER_KEY_TYPES = CodingHelper.encoderKeysForElements(
            GmlConstants.NS_GML_32, org.n52.sos.ogc.gml.time.Time.class, com.vividsolutions.jts.geom.Geometry.class,
            org.n52.sos.ogc.om.values.CategoryValue.class, org.n52.sos.ogc.gml.ReferenceType.class,
            org.n52.sos.ogc.om.values.QuantityValue.class, org.n52.sos.ogc.gml.CodeWithAuthority.class,
            org.n52.sos.ogc.gml.CodeType.class, AbstractSamplingFeature.class, SosEnvelope.class, FeatureCollection.class,
            AbstractGeometry.class, GenericMetaData.class, VerticalDatum.class, AbstractFeature.class,
            VerticalCRS.class, VerticalCS.class, CoordinateSystemAxis.class, DomainOfValidity.class);

    public GmlEncoderv321() {
        LOGGER.debug("Encoder for the following keys initialized successfully: {}!",
                Joiner.on(", ").join(ENCODER_KEY_TYPES));
    }

    @Override
    public Set<EncoderKey> getEncoderKeyType() {
        return Collections.unmodifiableSet(ENCODER_KEY_TYPES);
    }

    @Override
    public void addNamespacePrefixToMap(final Map<String, String> nameSpacePrefixMap) {
        nameSpacePrefixMap.put(GmlConstants.NS_GML_32, GmlConstants.NS_GML_PREFIX);
    }

    @Override
    public Set<SchemaLocation> getSchemaLocations() {
        return Sets.newHashSet(GmlConstants.GML_32_SCHEMAL_LOCATION);
    }

    @Override
    public XmlObject encode(final Object element, final Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        XmlObject encodedObject = null;
        if (element instanceof Time) {
            encodedObject = createTime((Time) element, additionalValues);
        } else if (element instanceof Geometry) {
            encodedObject = createPosition((Geometry) element, additionalValues);
        } else if (element instanceof CategoryValue) {
            encodedObject = createReferenceTypeForCategroyValue((CategoryValue) element);
        } else if (element instanceof org.n52.sos.ogc.gml.ReferenceType) {
            encodedObject = createReferencType((org.n52.sos.ogc.gml.ReferenceType) element);
        } else if (element instanceof CodeWithAuthority) {
            encodedObject = createCodeWithAuthorityType((CodeWithAuthority) element);
        } else if (element instanceof QuantityValue) {
            encodedObject = createMeasureType((QuantityValue) element);
        } else if (element instanceof org.n52.sos.ogc.gml.CodeType) {
            encodedObject = createCodeType((org.n52.sos.ogc.gml.CodeType) element);
        } else if (element instanceof AbstractFeature) {
            encodedObject = createFeaturePropertyType((AbstractFeature) element, additionalValues);
        } else if (element instanceof AbstractGeometry) {
            encodedObject = createGeomteryPropertyType((AbstractGeometry) element, additionalValues);
        } else if (element instanceof SosEnvelope) {
            encodedObject = createEnvelope((SosEnvelope) element, additionalValues);
        } else if (element instanceof GenericMetaData) {
            encodedObject = createGenericMetaData((GenericMetaData) element, additionalValues);
        } else if (element instanceof VerticalDatum) {
            encodedObject = createVerticalDatum((VerticalDatum) element, additionalValues);
        } else if (element instanceof DomainOfValidity) {
            encodedObject = createDomainOfValidity((DomainOfValidity) element, additionalValues);
        } else if (element instanceof VerticalCRS) {
            encodedObject = createVerticalCRS((VerticalCRS) element, additionalValues);
        } else if (element instanceof VerticalCS) {
            encodedObject = createVerticalCS((VerticalCS) element, additionalValues);
        } else if (element instanceof CoordinateSystemAxis) {
            encodedObject = createCoordinateSystemAxis((CoordinateSystemAxis) element, additionalValues);
        } else {
            throw new UnsupportedEncoderInputException(this, element);
        }
        // LOGGER.debug("Encoded object {} is valid: {}",
        // encodedObject.schemaType().toString(),
        // XmlHelper.validateDocument(encodedObject));
        return encodedObject;
    }

    protected XmlObject createFeaturePropertyType(final AbstractFeature feature,
            final Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        if (feature instanceof FeatureCollection) {
            return createFeatureCollection((FeatureCollection) feature, additionalValues);
        } else if (feature instanceof AbstractSamplingFeature) {
            return createFeature(feature, additionalValues);
        } else if (feature instanceof AbstractFeature && feature.isSetDefaultElementEncoding()) {
            return CodingHelper.encodeObjectToXml(feature.getDefaultElementEncoding(), feature);
        } else if (feature instanceof AbstractFeature && additionalValues.containsKey(HelperValues.ENCODE_NAMESPACE)) {
            return CodingHelper.encodeObjectToXml(additionalValues.get(HelperValues.ENCODE_NAMESPACE), feature,
                    additionalValues);
        } else {
            throw new UnsupportedEncoderInputException(this, feature);
        }
    }

    private XmlObject createFeatureCollection(final FeatureCollection element,
            final Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        final FeatureCollectionDocument featureCollectionDoc =
                FeatureCollectionDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        final FeatureCollectionType featureCollection = featureCollectionDoc.addNewFeatureCollection();
        featureCollection.setId(element.getGmlId());
        boolean document = checkAndPrepareHelperValues(additionalValues);
        if (element.isSetMembers()) {
            for (final AbstractFeature abstractFeature : element.getMembers().values()) {
                featureCollection.addNewFeatureMember().set(
                        createFeaturePropertyType(abstractFeature, additionalValues));
            }
        }
        if (document) {
            return featureCollectionDoc;
        }
        final FeaturePropertyType featurePropertyType =
                FeaturePropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        featurePropertyType.addNewAbstractFeature().set(featureCollection);
        return XmlHelper.substituteElement(featurePropertyType.getAbstractFeature(), featurePropertyType);
        // return featureCollection;
    }

    private boolean checkAndPrepareHelperValues(Map<HelperValues, String> additionalValues) {
        // add propertyType flag
        additionalValues.put(HelperValues.PROPERTY_TYPE, null);
        // check for document flag and remove flag from values if contained
        if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
            additionalValues.remove(HelperValues.DOCUMENT);
            return true;
        }
        return false;
    }

    private XmlObject createFeature(final AbstractFeature feature, final Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        final FeaturePropertyType featurePropertyType = createFeaturePropertyType();
        if (isNotSamplingFeature(feature) || additionalValues.containsKey(HelperValues.REFERENCED)) {
            featurePropertyType.setHref(feature.getIdentifierCodeWithAuthority().getValue());
            return featurePropertyType;
        } else {
            final AbstractSamplingFeature abstractSamplingFeature = (AbstractSamplingFeature) feature;
            if (feature.isSetGmlID()) {
                featurePropertyType.setHref("#" + feature.getGmlId());
                return featurePropertyType;
            } else {
                if (additionalValues.containsKey(HelperValues.ENCODE)
                        && additionalValues.get(HelperValues.ENCODE).equals("false") || !abstractSamplingFeature.isEncode()) {
                    featurePropertyType.setHref(feature.getIdentifierCodeWithAuthority().getValue());
                    if (feature instanceof AbstractSamplingFeature && abstractSamplingFeature.isSetName()) {
                        featurePropertyType.setTitle(abstractSamplingFeature.getFirstName().getValue());
                    }
                    return featurePropertyType;
                }
                if (!abstractSamplingFeature.isSetGeometry()) {
                    featurePropertyType.setHref(abstractSamplingFeature.getIdentifierCodeWithAuthority().getValue());
                    if (abstractSamplingFeature.isSetName()) {
                        featurePropertyType.setTitle(abstractSamplingFeature.getFirstName().getValue());
                    }
                    return featurePropertyType;
                }
                if (abstractSamplingFeature.isSetUrl()) {
                    featurePropertyType.setHref(abstractSamplingFeature.getUrl());
                    if (abstractSamplingFeature.isSetIdentifier()) {
                        featurePropertyType.setTitle(abstractSamplingFeature.getIdentifierCodeWithAuthority().getValue());
                    } else {
                        if (abstractSamplingFeature.isSetName()) {
                            featurePropertyType.setTitle(abstractSamplingFeature.getFirstName().getValue());
                        }
                    }
                    return featurePropertyType;
                } else {
                    String namespace;
                    if (additionalValues.containsKey(HelperValues.ENCODE_NAMESPACE)) {
                        namespace = additionalValues.get(HelperValues.ENCODE_NAMESPACE);
                    } else if (abstractSamplingFeature.isSetDefaultElementEncoding()) {
                        namespace = abstractSamplingFeature.getDefaultElementEncoding();
                    } else {
                        namespace = OMHelper.getNamespaceForFeatureType(abstractSamplingFeature.getFeatureType());
                    }
                    final XmlObject encodedXmlObject = CodingHelper.encodeObjectToXml(namespace, abstractSamplingFeature);

                    if (encodedXmlObject != null) {
                        return encodedXmlObject;
                    } else {
                        if (abstractSamplingFeature.getXmlDescription() != null) {
                            try {
                                // TODO how set gml:id in already existing
                                // XmlDescription? <-- XmlCursor
                                return XmlObject.Factory.parse(abstractSamplingFeature.getXmlDescription());
                            } catch (final XmlException xmle) {
                                throw new NoApplicableCodeException().causedBy(xmle).withMessage(
                                        "Error while encoding featurePropertyType!");
                            }
                        } else {
                            featurePropertyType.setHref(abstractSamplingFeature.getIdentifierCodeWithAuthority().getValue());
                            if (abstractSamplingFeature.isSetName()) {
                                featurePropertyType.setTitle(abstractSamplingFeature.getFirstName().getValue());
                            }
                            return featurePropertyType;
                        }
                    }
                }
            }
        }
    }

    @Override
    protected XmlObject createFeature(FeaturePropertyType featurePropertyType, AbstractFeature abstractFeature,
            Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        return featurePropertyType.set(createFeature(abstractFeature, additionalValues));
    }

    private boolean isNotSamplingFeature(final AbstractFeature feature) {
        return !(feature instanceof AbstractSamplingFeature);
    }

    private XmlObject createEnvelope(final SosEnvelope sosEnvelope, Map<HelperValues, String> additionalValues) {
        final int srid = sosEnvelope.getSrid();
        final EnvelopeType envelopeType = EnvelopeType.Factory.newInstance();
        final MinMax<String> minmax = SosHelper.getMinMaxFromEnvelope(sosEnvelope);
        envelopeType.addNewLowerCorner().setStringValue(minmax.getMinimum());
        envelopeType.addNewUpperCorner().setStringValue(minmax.getMaximum());
        envelopeType.setSrsName(getSrsName(srid));
        if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE) || additionalValues.containsKey(HelperValues.DOCUMENT)) {
            EnvelopeDocument doc = EnvelopeDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            doc.setEnvelope(envelopeType);
            return doc;
        }
        return envelopeType;
    }

    private XmlObject createTime(final Time time, final Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        if (time != null) {
            if (time instanceof TimeInstant) {
                if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
                    return createTimeInstantDocument((TimeInstant) time);
                } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
                    return createTimeInstantPropertyType((TimeInstant) time);
                } else {
                    return createTimeInstantType((TimeInstant) time, null);
                }
            } else if (time instanceof TimePeriod) {
                if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
                    return createTimePeriodDocument((TimePeriod) time);
                } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
                    return createTimePeriodPropertyType((TimePeriod) time);
                } else {
                    return createTimePeriodType((TimePeriod) time, null);
                }
            } else {
                throw new UnsupportedEncoderInputException(this, time);
            }
        }
        return null;
    }

    private XmlObject createTimePeriodDocument(final TimePeriod time) throws OwsExceptionReport {
        final TimePeriodDocument timePeriodDoc =
                TimePeriodDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        createTimePeriodType(time, timePeriodDoc.addNewTimePeriod());
        return timePeriodDoc;
    }

    private XmlObject createTimePeriodPropertyType(final TimePeriod time) throws OwsExceptionReport {
        final TimePeriodPropertyType timePeriodPropertyType =
                TimePeriodPropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        createTimePeriodType(time, timePeriodPropertyType.addNewTimePeriod());
        return timePeriodPropertyType;
    }

    /**
     * Creates a XML TimePeriod from the SOS time object.
     * 
     * @param timePeriod
     *            SOS time object
     * @param timePeriodType
     * @return XML TimePeriod
     * 
     * 
     * @throws OwsExceptionReport
     *             * if an error occurs.
     */
    private TimePeriodType createTimePeriodType(final TimePeriod timePeriod, TimePeriodType timePeriodType)
            throws OwsExceptionReport {
        if (timePeriodType == null) {
            timePeriodType = TimePeriodType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        }
        if (timePeriod.getGmlId() != null && !timePeriod.getGmlId().isEmpty()) {
            timePeriodType.setId(timePeriod.getGmlId());
        } else {
            timePeriodType.setId("tp_" + JavaHelper.generateID(timePeriod.toString() + System.currentTimeMillis()));
        }
        timePeriodType.setBeginPosition(createTimePositionType(timePeriod.getStartTimePosition()));
        timePeriodType.setEndPosition(createTimePositionType(timePeriod.getEndTimePosition()));

        return timePeriodType;
    }

    private XmlObject createTimeInstantDocument(final TimeInstant time) throws OwsExceptionReport {
        final TimeInstantDocument timeInstantDoc =
                TimeInstantDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        createTimeInstantType(time, timeInstantDoc.addNewTimeInstant());
        return timeInstantDoc;
    }

    private XmlObject createTimeInstantPropertyType(TimeInstant time) throws OwsExceptionReport {
        final TimeInstantPropertyType timeInstantPropertyType =
                TimeInstantPropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        createTimeInstantType(time, timeInstantPropertyType.addNewTimeInstant());
        return timeInstantPropertyType;
    }

    /**
     * Creates a XML TimeInstant from the SOS time object.
     * 
     * @param timeInstant
     *            SOS time object
     * @param timeInstantType
     * @return XML TimeInstant
     * 
     * 
     * @throws OwsExceptionReport
     *             * if an error occurs.
     */
    private TimeInstantType createTimeInstantType(final TimeInstant timeInstant, TimeInstantType timeInstantType)
            throws OwsExceptionReport {
        // create time instant
        if (timeInstantType == null) {
            timeInstantType = TimeInstantType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        }
        if (timeInstant.isSetGmlId()) {
            timeInstantType.setId(timeInstant.getGmlId());
        } else {
            timeInstantType.setId("ti_"
                    + JavaHelper.generateID(timeInstantType.toString() + System.currentTimeMillis()));
        }
        timeInstantType.setTimePosition(createTimePositionType(timeInstant.getTimePosition()));
        return timeInstantType;
    }

    private TimePositionType createTimePositionType(final TimePosition timePosition) throws DateTimeFormatException {
        final TimePositionType xbTimePosition = TimePositionType.Factory.newInstance();
        if (!timePosition.isSetTime()) {
            if (timePosition.isSetIndeterminateValue()) {
                xbTimePosition.setIndeterminatePosition(TimeIndeterminateValueType.Enum.forString(timePosition
                        .getIndeterminateValue().name()));
            } else {
                xbTimePosition.setIndeterminatePosition(TimeIndeterminateValueType.Enum
                        .forString(TimeIndeterminateValue.unknown.name()));
            }
        } else {
            final String endString = DateTimeHelper.formatDateTime2String(timePosition);

            // concat minutes for timeZone offset, because gml requires
            // xs:dateTime, which needs minutes in
            // timezone offset
            // TODO enable really
            xbTimePosition.setStringValue(endString);
        }
        return xbTimePosition;
    }

    private XmlObject createGeomteryPropertyType(AbstractGeometry element, Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        GeometryPropertyType geometryPropertyType = GeometryPropertyType.Factory.newInstance();
        if (element.isReferenced()) {
            geometryPropertyType.setHref(element.getGmlId());
        } else {
            AbstractGeometryType xmlObject = createAbstractGeometry(element, additionalValues);
            geometryPropertyType.setAbstractGeometry(xmlObject);
            XmlHelper.substituteElement(geometryPropertyType.getAbstractGeometry(), xmlObject);
        }

        return geometryPropertyType;
    }

    private AbstractGeometryType createAbstractGeometry(AbstractGeometry element,
            Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        if (element.isSetGmlID() && !additionalValues.containsKey(HelperValues.GMLID)) {
            additionalValues.put(HelperValues.GMLID, element.getGmlId());
        }
        XmlObject xbGeometry = createPosition(element.getGeometry(), additionalValues);
        AbstractGeometryType abstractGeometryType = null;
        if (xbGeometry instanceof AbstractGeometryType) {
            abstractGeometryType = (AbstractGeometryType) xbGeometry;
        } else if (xbGeometry instanceof GeometryPropertyType) {
            abstractGeometryType = ((GeometryPropertyType) xbGeometry).getAbstractGeometry();
        } else {
            throw new UnsupportedEncoderInputException(this, element);
        }

        if (element.isSetIdentifier()) {
            abstractGeometryType.setIdentifier(createCodeWithAuthorityType(element.getIdentifierCodeWithAuthority()));
        }
        if (element.isSetName()) {
            for (org.n52.sos.ogc.gml.CodeType codeType : element.getName()) {
                abstractGeometryType.addNewName().set(createCodeType(codeType));
            }
        }
        if (element.isSetDescription()) {
            abstractGeometryType.addNewDescription().setStringValue(element.getDescription());
        }
        return abstractGeometryType;
    }

    private XmlObject createPosition(Geometry geom, Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        String gmlId = additionalValues.get(HelperValues.GMLID);
        if (geom instanceof Point) {
            final PointType xbPoint = PointType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            xbPoint.setId(getGmlID(geom, gmlId));
            createPointFromJtsGeometry((Point) geom, xbPoint);
            if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
                PointDocument xbPointDoc =
                        PointDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                xbPointDoc.setPoint(xbPoint);
                return xbPointDoc;
            } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
                GeometryPropertyType geometryPropertyType =
                        GeometryPropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                geometryPropertyType.setAbstractGeometry(xbPoint);
                geometryPropertyType.getAbstractGeometry().substitute(GmlConstants.QN_POINT_32, PointType.type);
                return geometryPropertyType;
            }
            return xbPoint;
        }

        else if (geom instanceof LineString) {
            final LineStringType xbLineString =
                    LineStringType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            xbLineString.setId(getGmlID(geom, gmlId));
            createLineStringFromJtsGeometry((LineString) geom, xbLineString);
            if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
                LineStringDocument xbLineStringDoc =
                        LineStringDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                xbLineStringDoc.setLineString(xbLineString);
                return xbLineStringDoc;
            } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
                GeometryPropertyType geometryPropertyType =
                        GeometryPropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                geometryPropertyType.setAbstractGeometry(xbLineString);
                geometryPropertyType.getAbstractGeometry().substitute(GmlConstants.QN_LINESTRING_32,
                        LineStringType.type);
                return geometryPropertyType;
            }
            return xbLineString;
        }

        else if (geom instanceof Polygon) {
            final PolygonType xbPolygon =
                    PolygonType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            xbPolygon.setId(getGmlID(geom, gmlId));
            createPolygonFromJtsGeometry((Polygon) geom, xbPolygon);
            if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
                PolygonDocument xbPolygonDoc =
                        PolygonDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                xbPolygonDoc.setPolygon(xbPolygon);
                return xbPolygonDoc;
            } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
                GeometryPropertyType geometryPropertyType =
                        GeometryPropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                geometryPropertyType.setAbstractGeometry(xbPolygon);
                geometryPropertyType.getAbstractGeometry().substitute(GmlConstants.QN_POLYGON_32, PolygonType.type);
                return geometryPropertyType;
            }
            return xbPolygon;
        } 
        
        else if (geom instanceof MultiPoint) {
            final MultiPointType xbMultiPoint = MultiPointType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
            String id = getGmlID(geom, gmlId);
            xbMultiPoint.setId(id);
            createMultiPointFromJtsGeometry((MultiPoint) geom, xbMultiPoint, id);
            
            if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
                MultiPointDocument xbMultiPointDoc =
                        MultiPointDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                xbMultiPointDoc.setMultiPoint(xbMultiPoint);
                return xbMultiPointDoc;
            } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
                GeometryPropertyType geometryPropertyType =
                        GeometryPropertyType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
                geometryPropertyType.setAbstractGeometry(xbMultiPoint);
                geometryPropertyType.getAbstractGeometry().substitute(GmlConstants.QN_MULTI_POINT_32, PolygonType.type);
                return geometryPropertyType;
            }
            return xbMultiPoint;
        }
        
        else {
            throw new UnsupportedEncoderInputException(this, geom);
        }
    }

    @SuppressWarnings("rawtypes")
    private String getGmlID(Geometry geom, String gmlId) {
        String id = null;
        if (!Strings.isNullOrEmpty(gmlId)) {
            id = gmlId;
        } else if (geom.getUserData() != null && geom.getUserData() instanceof Map
                && ((Map) geom.getUserData()).containsKey(HelperValues.GMLID.name())) {
            id = (String) ((Map) geom.getUserData()).get(HelperValues.GMLID.name());
        } else {
            id = JavaHelper.generateID(geom.toText());
        }
        return geom.getGeometryType() + "_" + id;
    }

    /**
     * Creates a XML Point from a SOS Point.
     * 
     * @param jtsPoint
     *            SOS Point
     * @param xbPoint
     *            XML Point
     */
    private void createPointFromJtsGeometry(final Point jtsPoint, final PointType xbPoint) throws OwsExceptionReport {
        final DirectPositionType xbPos = xbPoint.addNewPos();
        xbPos.setSrsName(getSrsName(jtsPoint));
        xbPos.setStringValue(JTSHelper.getCoordinatesString(jtsPoint));
    }

    /**
     * Creates a XML LineString from a SOS LineString.
     * 
     * @param jtsLineString
     *            SOS LineString
     * @param xbLst
     *            XML LinetSring
     */
    private void createLineStringFromJtsGeometry(final LineString jtsLineString, final LineStringType xbLst)
            throws OwsExceptionReport {
        final String srsName = getSrsName(jtsLineString);
        xbLst.setSrsName(srsName);
        final DirectPositionListType xbPosList = xbLst.addNewPosList();
        xbPosList.setSrsName(srsName);
        xbPosList.setStringValue(JTSHelper.getCoordinatesString(jtsLineString));

    }

    /**
     * Creates a XML Polygon from a SOS Polygon.
     * 
     * @param jtsPolygon
     *            SOS Polygon
     * @param xbPolType
     *            XML Polygon
     */
    private void createPolygonFromJtsGeometry(final Polygon jtsPolygon, final PolygonType xbPolType)
            throws OwsExceptionReport {
        final List<?> jtsPolygons = PolygonExtracter.getPolygons(jtsPolygon);
        final String srsName = getSrsName(jtsPolygon);

        for (int i = 0; i < jtsPolygons.size(); i++) {

            final Polygon pol = (Polygon) jtsPolygons.get(i);

            AbstractRingPropertyType xbArpt = xbPolType.addNewExterior();
            AbstractRingType xbArt = xbArpt.addNewAbstractRing();

            LinearRingType xbLrt = LinearRingType.Factory.newInstance();

            // Exterior ring
            LineString ring = pol.getExteriorRing();
            DirectPositionListType xbPosList = xbLrt.addNewPosList();

            xbPosList.setSrsName(srsName);
            xbPosList.setStringValue(JTSHelper.getCoordinatesString(ring));
            xbArt.set(xbLrt);

            // Rename element name for output
            XmlCursor cursor = xbArpt.newCursor();
            if (cursor.toChild(GmlConstants.QN_ABSTRACT_RING_32)) {
                cursor.setName(GmlConstants.QN_LINEAR_RING_32);
            }

            // Interior ring
            final int numberOfInteriorRings = pol.getNumInteriorRing();
            for (int ringNumber = 0; ringNumber < numberOfInteriorRings; ringNumber++) {
                xbArpt = xbPolType.addNewInterior();
                xbArt = xbArpt.addNewAbstractRing();

                xbLrt = LinearRingType.Factory.newInstance();

                ring = pol.getInteriorRingN(ringNumber);

                xbPosList = xbLrt.addNewPosList();
                xbPosList.setSrsName(srsName);
                xbPosList.setStringValue(JTSHelper.getCoordinatesString(ring));
                xbArt.set(xbLrt);

                // Rename element name for output
                cursor = xbArpt.newCursor();
                if (cursor.toChild(GmlConstants.QN_ABSTRACT_RING_32)) {
                    cursor.setName(GmlConstants.QN_LINEAR_RING_32);
                }
            }
        }
    }

    private void createMultiPointFromJtsGeometry(MultiPoint geom, MultiPointType xbMultiPoint, String id) throws OwsExceptionReport {
        for (int i = 0; i < geom.getNumGeometries(); i++) {
            Geometry geometry = geom.getGeometryN(i);
            if (geometry instanceof Point) {
                PointType pt = xbMultiPoint.addNewPointMember().addNewPoint();
                pt.setId(id + "_" + i);
                createPointFromJtsGeometry((Point)geometry, pt);
            }
        }
    }

    private XmlObject createReferenceTypeForCategroyValue(final CategoryValue categoryValue) {
        final ReferenceType xbRef = ReferenceType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        if (categoryValue.isSetValue()) {
            if (categoryValue.getValue().startsWith("http://")) {
                xbRef.setHref(categoryValue.getValue());
            } else {
                xbRef.setTitle(categoryValue.getValue());
            }
            if (categoryValue.isSetUnit()) {
                xbRef.setRole(categoryValue.getUnit());
            }
        } else {
            xbRef.setNil();
        }
        return xbRef;
    }

    private ReferenceType createReferencType(final org.n52.sos.ogc.gml.ReferenceType sosReferenceType) {
        final ReferenceType referenceType = ReferenceType.Factory.newInstance();
        if (sosReferenceType.isEmpty()) {
            referenceType.setNilReason(GmlConstants.NilReason.unknown.name());
        } else {
            if (sosReferenceType.isSetHref()) {
                referenceType.setHref(sosReferenceType.getHref());
            }
            if (sosReferenceType.isSetTitle()) {
                referenceType.setTitle(sosReferenceType.getTitle());
            }
            if (sosReferenceType.isSetRole()) {
                referenceType.setRole(sosReferenceType.getRole());
            }
        }
        return referenceType;
    }

    private CodeWithAuthorityType createCodeWithAuthorityType(final CodeWithAuthority sosCodeWithAuthority) {
        if (!sosCodeWithAuthority.isSetValue()) {
            final String exceptionText =
                    String.format("The required 'value' parameter is empty for encoding %s!",
                            CodeWithAuthorityType.class.getName());
            LOGGER.error(exceptionText);
            throw new IllegalArgumentException(exceptionText);
        }
        final CodeWithAuthorityType codeWithAuthority =
                CodeWithAuthorityType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        codeWithAuthority.setStringValue(sosCodeWithAuthority.getValue());
        if (sosCodeWithAuthority.isSetCodeSpace()) {
            codeWithAuthority.setCodeSpace(sosCodeWithAuthority.getCodeSpace());
        } else {
            codeWithAuthority.setCodeSpace(OGCConstants.UNKNOWN);
        }
        return codeWithAuthority;
    }

    private CodeType createCodeType(final org.n52.sos.ogc.gml.CodeType sosCodeType) {
        if (!sosCodeType.isSetValue()) {
            final String exceptionText =
                    String.format("The required 'value' parameter is empty for encoding %s!", CodeType.class.getName());
            LOGGER.error(exceptionText);
            throw new IllegalArgumentException(exceptionText);
        }
        final CodeType codeType = CodeType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        codeType.setStringValue(sosCodeType.getValue());
        if (sosCodeType.isSetCodeSpace()) {
            codeType.setCodeSpace(sosCodeType.getCodeSpace());
        } else {
            codeType.setCodeSpace(OGCConstants.UNKNOWN);
        }
        return codeType;
    }
    
    private XmlObject createGenericMetaData(GenericMetaData element, Map<HelperValues, String> additionalValues)
            throws OwsExceptionReport {
        GenericMetaDataDocument gmdd = GenericMetaDataDocument.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        GenericMetaDataType gmdt = gmdd.addNewGenericMetaData();
        if (element.getContent() instanceof DefaultEncoding && ((DefaultEncoding)element.getContent()).isSetDefaultElementEncoding()) {
            Map<HelperValues, String> helperValues = new EnumMap<HelperValues, String>(HelperValues.class);
            // TODO check
            helperValues.put(HelperValues.PROPERTY_TYPE, "true");
            gmdt.set(CodingHelper.encodeObjectToXml(
                    ((DefaultEncoding) element.getContent()).getDefaultElementEncoding(), element.getContent(),
                    helperValues));
        }
        if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
            return gmdd;
        }
        return gmdt;
    }

    protected MeasureType createMeasureType(final QuantityValue quantityValue) throws OwsExceptionReport {
        if (!quantityValue.isSetValue()) {
            throw new NoApplicableCodeException().withMessage(
                    "The required 'value' parameter is empty for encoding %s!", MeasureType.class.getName());
        }
        final MeasureType measureType =
                MeasureType.Factory.newInstance(XmlOptionsHelper.getInstance().getXmlOptions());
        measureType.setDoubleValue(quantityValue.getValue().doubleValue());
        if (quantityValue.isSetUnit()) {
            measureType.setUom(quantityValue.getUnit());
        } else {
            measureType.setUom(OGCConstants.UNKNOWN);
        }
    
        return measureType;
    }

    private XmlObject createVerticalDatum(VerticalDatum verticalDatum, Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        VerticalDatumType vdt = VerticalDatumType.Factory.newInstance();
        addAbstractDatumValues(vdt, verticalDatum, additionalValues);
        if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
            VerticalDatumDocument vdd = VerticalDatumDocument.Factory.newInstance();
            VerticalDatumPropertyType vdpt = VerticalDatumPropertyType.Factory.newInstance();
            vdpt.setVerticalDatum(vdt);
            vdd.setVerticalDatum(vdpt);
            return vdd;
        } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
            VerticalDatumPropertyType vdpt = VerticalDatumPropertyType.Factory.newInstance();
            vdpt.setVerticalDatum(vdt);
            return vdpt;
        }
        return vdt;
    }
    
    private void addAbstractDatumValues(AbstractDatumType adt, AbstractDatum abstractDatum, Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        addDefinitonValues(adt, abstractDatum);
        if (abstractDatum.hasAnchorDefinition()) {
            adt.setAnchorDefinition(createCodeType(abstractDatum.getAnchorDefinition()));
        } 
        if (abstractDatum.hasDomainOfValidity()) {
            net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity dov = adt.addNewDomainOfValidity();
            Referenceable<DomainOfValidity> domainOfValidity = abstractDatum.getDomainOfValidity();
            if (domainOfValidity.isReference()) {
                Reference reference = domainOfValidity.getReference();
                if (reference.getActuate().isPresent()) {
                    dov.setActuate(ActuateType.Enum.forString(reference.getActuate().get()));
                }
                if (reference.getArcrole().isPresent()) {
                    dov.setHref(reference.getArcrole().get());
                }
                if (reference.getHref().isPresent()) {
                    dov.setHref(reference.getHref().get().toString());
                }
                if (reference.getRole().isPresent()) {
                    dov.setRole(reference.getRole().get());
                }
                if (reference.getShow().isPresent()) {
                    dov.setShow(ShowType.Enum.forString(reference.getShow().get()));
                }
                if (reference.getTitle().isPresent()) {
                    dov.setTitle(reference.getTitle().get());
                }
                if (reference.getType().isPresent()) {
                    dov.setType(TypeType.Enum.forString(reference.getType().get()));
                }
            } else { 
                if (domainOfValidity.isInstance()) {
                    Nillable<DomainOfValidity> nillable = domainOfValidity.getInstance();
                    if (nillable.isPresent()) {
                        net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity xml = createDomainOfValidity(
                                nillable.get(), new EnumMap<HelperValues, String>(HelperValues.class));
                        if (xml != null) {
                            dov.set(xml);
                        } else {
                            dov.setNil();
                            dov.setNilReason(Nillable.missing().get());
                        }
                    } else {
                        dov.setNil();
                        if (nillable.hasReason()) {
                            dov.setNilReason(nillable.getNilReason().get());
                        } else {
                            dov.setNilReason(Nillable.missing().get());
                        }
                    }
                }
            }
        } 
        if (abstractDatum.hasRealizationEpoch()) {
            abstractDatum.getRealizationEpoch();
            adt.setRealizationEpoch(abstractDatum.getRealizationEpoch().toCalendar(Locale.ROOT));
        } 
        for (String scope : abstractDatum.getScope()) {
            adt.addNewScope().setStringValue(scope);
        }
    }

    private net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity createDomainOfValidity(DomainOfValidity domainOfValidity, Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity dov = net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity.Factory.newInstance();
        if (domainOfValidity.hasExExtent()) {
            EXExtentType exet = dov.addNewEXExtent();
            XmlObject xml = CodingHelper.encodeObjectToXml(domainOfValidity.getExExtent().getDefaultElementEncoding(), domainOfValidity.getExExtent());
            if (xml != null) {
                exet.set(xml);
            }
        }
        return dov;
    }

    private XmlObject createVerticalCRS(VerticalCRS verticalCRS, Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        VerticalCRSType vcrst = VerticalCRSType.Factory.newInstance();
        addAbstractCRSValues(vcrst, verticalCRS);
        // verticalCS
        Referenceable<VerticalCS> verticalCS = verticalCRS.getVerticalCS();
        VerticalCSPropertyType vcspt = vcrst.addNewVerticalCS();
        if (verticalCS.isReference()) {
            Reference reference = verticalCS.getReference();
            if (reference.getActuate().isPresent()) {
                vcspt.setActuate(ActuateType.Enum.forString(reference.getActuate().get()));
            }
            if (reference.getArcrole().isPresent()) {
                vcspt.setHref(reference.getArcrole().get());
            }
            if (reference.getHref().isPresent()) {
                vcspt.setHref(reference.getHref().get().toString());
            }
            if (reference.getRole().isPresent()) {
                vcspt.setRole(reference.getRole().get());
            }
            if (reference.getShow().isPresent()) {
                vcspt.setShow(ShowType.Enum.forString(reference.getShow().get()));
            }
            if (reference.getTitle().isPresent()) {
                vcspt.setTitle(reference.getTitle().get());
            }
            if (reference.getType().isPresent()) {
                vcspt.setType(TypeType.Enum.forString(reference.getType().get()));
            }
        } else {
            if (verticalCS.isInstance()) {
                Nillable<VerticalCS> nillable = verticalCS.getInstance();
                if (nillable.isPresent()) {
                    XmlObject xml = createVerticalCS(nillable.get(), new EnumMap<HelperValues, String>(HelperValues.class));
                    if (xml != null && xml instanceof VerticalCSType) {
                        vcspt.set((VerticalCSType) xml);
                    } else {
                        vcspt.setNil();
                        vcspt.setNilReason(Nillable.missing().get());
                    }
                } else {
                    vcspt.setNil();
                    if (nillable.hasReason()) {
                        vcspt.setNilReason(nillable.getNilReason().get());
                    } else {
                        vcspt.setNilReason(Nillable.missing().get());
                    }
                }
            }
        }
        // verticalDatum
        Referenceable<VerticalDatum> verticalDatum = verticalCRS.getVerticalDatum();
        VerticalDatumPropertyType vdpt = vcrst.addNewVerticalDatum();
        if (verticalDatum.isReference()) {
            Reference reference = verticalDatum.getReference();
            if (reference.getActuate().isPresent()) {
                vdpt.setActuate(ActuateType.Enum.forString(reference.getActuate().get()));
            }
            if (reference.getArcrole().isPresent()) {
                vdpt.setHref(reference.getArcrole().get());
            }
            if (reference.getHref().isPresent()) {
                vdpt.setHref(reference.getHref().get().toString());
            }
            if (reference.getRole().isPresent()) {
                vdpt.setRole(reference.getRole().get());
            }
            if (reference.getShow().isPresent()) {
                vdpt.setShow(ShowType.Enum.forString(reference.getShow().get()));
            }
            if (reference.getTitle().isPresent()) {
                vdpt.setTitle(reference.getTitle().get());
            }
            if (reference.getType().isPresent()) {
                vdpt.setType(TypeType.Enum.forString(reference.getType().get()));
            }
        } else {
            if (verticalDatum.isInstance()) {
                Nillable<VerticalDatum> nillable = verticalDatum.getInstance();
                if (nillable.isPresent()) {
                    XmlObject xml = createVerticalDatum(nillable.get(), new EnumMap<HelperValues, String>(HelperValues.class));
                    if (xml != null && xml instanceof VerticalDatumType) {
                        vdpt.setVerticalDatum((VerticalDatumType) xml);
                    } else {
                        vdpt.setNil();
                        vdpt.setNilReason(Nillable.missing().get());
                    }
                } else {
                    vdpt.setNil();
                    if (nillable.hasReason()) {
                        vdpt.setNilReason(nillable.getNilReason().get());
                    } else {
                        vdpt.setNilReason(Nillable.missing().get());
                    }
                }
            }
        }
        if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
            VerticalCRSPropertyType vcrspt = VerticalCRSPropertyType.Factory.newInstance();
            vcrspt.setVerticalCRS(vcrst);
            return vcrspt;
        }
        return vcrst;
    }
    
    private void addDefinitonValues(DefinitionType dt, Definition definition) throws OwsExceptionReport {
        if (!definition.isSetGmlID()) {
            definition.setGmlId("id_" + JavaHelper.generateID(definition.getGmlId()));
        }
        dt.setId(definition.getGmlId());
        if (!addIdentifier(dt, definition)) {
            dt.setIdentifier(createCodeWithAuthorityType(new CodeWithAuthority(JavaHelper.generateID(definition.toString()))));
        }
        addName(dt, definition);
        addDescription(dt, definition);
        if (definition.hasRemarks()) {
            dt.setRemarks(definition.getRemarks());
        } 
    }

    private void addAbstractCRSValues(AbstractCRSType acrst, AbstractCRS abstractCRS) throws OwsExceptionReport {
        addDefinitonValues(acrst, abstractCRS);
        if (abstractCRS.hasDomainOfValidity()) {
            for (Referenceable<DomainOfValidity> domainOfValidity : abstractCRS.getDomainOfValidity()) {
                net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity dov = acrst.addNewDomainOfValidity();
                if (domainOfValidity.isReference()) {
                    Reference reference = domainOfValidity.getReference();
                    if (reference.getActuate().isPresent()) {
                        dov.setActuate(ActuateType.Enum.forString(reference.getActuate().get()));
                    }
                    if (reference.getArcrole().isPresent()) {
                        dov.setHref(reference.getArcrole().get());
                    }
                    if (reference.getHref().isPresent()) {
                        dov.setHref(reference.getHref().get().toString());
                    }
                    if (reference.getRole().isPresent()) {
                        dov.setRole(reference.getRole().get());
                    }
                    if (reference.getShow().isPresent()) {
                        dov.setShow(ShowType.Enum.forString(reference.getShow().get()));
                    }
                    if (reference.getTitle().isPresent()) {
                        dov.setTitle(reference.getTitle().get());
                    }
                    if (reference.getType().isPresent()) {
                        dov.setType(TypeType.Enum.forString(reference.getType().get()));
                    }
                } else {
                    if (domainOfValidity.isInstance()) {
                        Nillable<DomainOfValidity> nillable = domainOfValidity.getInstance();
                        if (nillable.isPresent()) {
                            net.opengis.gml.x32.DomainOfValidityDocument.DomainOfValidity xml = createDomainOfValidity(
                                    nillable.get(), new EnumMap<HelperValues, String>(HelperValues.class));
                            if (xml != null) {
                                dov.set(xml);
                            } else {
                                dov.setNil();
                                dov.setNilReason(Nillable.missing().get());
                            }
                        } else {
                            dov.setNil();
                            if (nillable.hasReason()) {
                                dov.setNilReason(nillable.getNilReason().get());
                            } else {
                                dov.setNilReason(Nillable.missing().get());
                            }
                        }
                    }
                }
            }
        } 
        if (abstractCRS.hasScope()) {
            for (String scope : abstractCRS.getScope()) {
                acrst.addNewScope().setStringValue(scope);
            }
        }
    }

    private XmlObject createVerticalCS(VerticalCS verticalCS, Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        VerticalCSType vcst = VerticalCSType.Factory.newInstance();
        addAbstractCoordincateSystemValues(vcst, verticalCS);
        if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
            VerticalCSDocument vcsd = VerticalCSDocument.Factory.newInstance();
            VerticalCSPropertyType vcdpt = VerticalCSPropertyType.Factory.newInstance();
            vcdpt.setVerticalCS(vcst);
            vcsd.setVerticalCS(vcdpt);
            return vcsd;
        } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
            VerticalCSPropertyType vcdpt = VerticalCSPropertyType.Factory.newInstance();
            vcdpt.setVerticalCS(vcst);
            return vcdpt;
        }
        return vcst;
    }

    private void addAbstractCoordincateSystemValues(AbstractCoordinateSystemType acst, AbstractCoordinateSystem abstractCoordinateSystem) throws OwsExceptionReport {
        addDefinitonValues(acst, abstractCoordinateSystem);
        acst.setAggregationType(AggregationType.Enum.forString(abstractCoordinateSystem.getAggregation().name()));
        for (Referenceable<CoordinateSystemAxis> coordinateSystemAxis : abstractCoordinateSystem.getCoordinateSystemAxis()) {
            CoordinateSystemAxisPropertyType csapt = acst.addNewAxis();
            if (coordinateSystemAxis.isReference()) {
                Reference reference = coordinateSystemAxis.getReference();
                if (reference.getActuate().isPresent()) {
                    csapt.setActuate(ActuateType.Enum.forString(reference.getActuate().get()));
                }
                if (reference.getArcrole().isPresent()) {
                    csapt.setHref(reference.getArcrole().get());
                }
                if (reference.getHref().isPresent()) {
                    csapt.setHref(reference.getHref().get().toString());
                }
                if (reference.getRole().isPresent()) {
                    csapt.setRole(reference.getRole().get());
                }
                if (reference.getShow().isPresent()) {
                    csapt.setShow(ShowType.Enum.forString(reference.getShow().get()));
                }
                if (reference.getTitle().isPresent()) {
                    csapt.setTitle(reference.getTitle().get());
                }
                if (reference.getType().isPresent()) {
                    csapt.setType(TypeType.Enum.forString(reference.getType().get()));
                }
            } else {
                if (coordinateSystemAxis.isInstance()) {
                    Nillable<CoordinateSystemAxis> nillable = coordinateSystemAxis.getInstance();
                    if (nillable.isPresent()) {
                        XmlObject xml = createCoordinateSystemAxis(nillable.get(), new EnumMap<HelperValues, String>(HelperValues.class));
                        if (xml != null && xml instanceof CoordinateSystemAxisType) {
                            csapt.addNewCoordinateSystemAxis().set(xml);
                        } else {
                            csapt.setNil();
                            csapt.setNilReason(Nillable.missing().get());
                        }
                    } else {
                        csapt.setNil();
                        if (nillable.hasReason()) {
                            csapt.setNilReason(nillable.getNilReason().get());
                        } else {
                            csapt.setNilReason(Nillable.missing().get());
                        }
                    }
                }
            }
        }
    }

    private XmlObject createCoordinateSystemAxis(CoordinateSystemAxis coordinateSystemAxis,
            Map<HelperValues, String> additionalValues) throws OwsExceptionReport {
        CoordinateSystemAxisType csat = CoordinateSystemAxisType.Factory.newInstance();
        addDefinitonValues(csat, coordinateSystemAxis);
        csat.setAxisAbbrev(createCodeType(coordinateSystemAxis.getAxisAbbrev()));
        csat.setAxisDirection(createCodeWithAuthorityType(coordinateSystemAxis.getAxisDirection()));
        if (coordinateSystemAxis.isSetMinimumValue()) {
            csat.setMinimumValue(coordinateSystemAxis.getMinimumValue());
        }
        if (coordinateSystemAxis.isSetMaximumValue()) {
            csat.setMaximumValue(coordinateSystemAxis.getMaximumValue());
        }
        if (coordinateSystemAxis.isSetRangeMeaning()) {
            csat.setRangeMeaning(createCodeWithAuthorityType(coordinateSystemAxis.getRangeMeaning()));
        }
        csat.setUom(coordinateSystemAxis.getUom());
        if (additionalValues.containsKey(HelperValues.DOCUMENT)) {
            CoordinateSystemAxisDocument csad = CoordinateSystemAxisDocument.Factory.newInstance();
            csad.setCoordinateSystemAxis(csat);
            return csad;
        } else if (additionalValues.containsKey(HelperValues.PROPERTY_TYPE)) {
            CoordinateSystemAxisPropertyType csapt = CoordinateSystemAxisPropertyType.Factory.newInstance();
            csapt.setCoordinateSystemAxis(csat);
            return csapt;
        }
        return csat;
    }

    protected String getSrsName(final Geometry geom) {
        return getSrsName(geom.getSRID());
    }

    protected String getSrsName(final int srid) {
        return ServiceConfiguration.getInstance().getSrsNamePrefixSosV2() + srid;
    }
}
